#define VERSION			"GSModeSelector_v0.23e"
#define AUTHORS			"doctorxyz and dlanor"
#define CNF_VERSION	"v0.22" //GSM version which defined current CNF format
/*
# GS Mode Selector - Force (set and keep) a GS Mode, then load & exec a PS2 ELF
#------------------------------------------------------------------------------
# Copyright 2009, doctorxyz & dlanor
# Licenced under Academic Free License version 2.0
# Review LICENSE file for further details.
*
* Thank goes to (in alphabetical order)...
* bootsector         * goriath         * rikimaru    
* dariuszg           * guxtabo         * rodpad       
* darthkamikaz3      * jepjepjep       * stefanol69   
* david_wgtn         * jimmikaelkael   * tna          
* don roberto        * katananja       * urbigbro     
* euzeb2             * kevstah2004     * vegeta       
* ffgriever          * maverickhl      * zin0099      
* gilberto moreira   * Neme
* ... and all those that contributed (and still contribute) for this project!
*
*/

#define BOOT_ELF	"BOOT/BOOT.ELF"			//ELF to boot on exit
#define HDL_ELF		"BOOT/HDLOADER.ELF"	//Alternate ELF to boot on exit

#include <stdio.h>
#include <kernel.h>
#include <sifrpc.h>
#include <libhdd.h>
#include <debug.h>
#include <malloc.h>
#include <fileXio_rpc.h>
#include <loadfile.h>
#include <fileio.h>
#include <string.h>
#include <libmc.h>
#include <hw.h>
#include <libpad.h>

#include <sys/fcntl.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include "modules.h"
#include "timer.h"
#include "pad.h"

#include "gsKit.h"

typedef struct gsm_vmode {
	u32	interlace;
	u32	mode;
	u32	field;
	u64	display;
	u64	syncv;
	u64 smode2;
} GSM_vmode;

u32	user_vmode_slots = 16;	//Must be a power of two (uses bitmask = value-1)

GSM_vmode user_vmode[16] = {
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0}
};

struct gsm_settings {
	char	GSM_CNF_version[64];
	char  option_1[64];
	int   option_2;
};

int	MC_index = -1;	//index of MC where a CNF was successfully found or created

struct gsm_settings *GSM = NULL;

char ROMVER_data[20]; 	//16 byte file read from rom0:ROMVER at init

/* Prototypes for External Functions */
void RunLoaderElf(char *filename, char *);

#define MAKE_J(func)		(u32)( (0x02 << 26) | (((u32)func) / 4) )	// Jump (MIPS instruction)
#define NOP					0x00000000									// No Operation (MIPS instruction)

#define MAX_PATH 260


// TSR - Data Section
// ------------------

	#include "KSEG_Macros.h"

	KDef(void, KSeg, 0x00080000);  //KSeg is the base of our KSeg area
	KRel(u128, MIPS_Regs, 0x0000); //Here we save exception entry registers
	//NB: The MIPS_Regs area must be 16-byte aligned, for 128 bit MIPS registers
	//But our coding also requires MIPS_Regs == KSeg, with the lower 16 bits zero

	KRel(u32, Target_Interlace, 0x0200);
	KRel(u32, Target_Mode,      0x0204);
	KRel(u32, Target_Field,     0x0208);

	KRel(u32, Adapt_Flags,  0x020C);	//Four Byte flags combined
	KRel(u8,  Adapt_Flag_0, 0x020C);	//Byte flag
	KRel(u8,  Adapt_Flag_1, 0x020D);	//Byte flag
	KRel(u8,  Adapt_Flag_2, 0x020E);	//Byte flag
	KRel(u8,  Adapt_Flag_3, 0x020F);	//Byte flag

	KRel(u64, Target_Display, 0x0210);

	KRel(u32,  Source_Interlace, 0x0220);
	KRel(u32,  Source_Mode,    0x0224);
	KRel(u32,  Source_Field,   0x0228);
	KRel(u64, Source_Display, 0x0230);
//NB: Source_Display above is unsafe! Something zeroes it...
//NB: For this reason we now use Safe_Orig_display instead
	KRel(void*, syscallTable, 0x0240);  //Pointer to function vector table
	KRel(void*, sceSetGsCrt, 0x0244);	  //Pointer to function

	KRel(u64, Target_SMode2,	0x0248);
	KRel(u64, Target_SyncV,   0x0250);
	KRel(u64, Source_SMode2, 0x0258);
	KRel(u64, Source_SyncV, 0x0260);

	KRel(void*, DisplayHandler_p, 0x0268); //Allows DisplayHandler to find itself

	KRel(u8, Adapt_DoubleHeight, 0x026C); //Flags Height Doubling at INT&FFMD
	KRel(u8, Adapt_SMode2, 0x026D);				//Adapted SMode2 patch value

	KRel(u64, Safe_Orig_Display, 0x0270);
	KRel(u64, Calc_Display, 0x0278);

	KDef(void, KSEG_func_base, 0x00080280);

	KDef(u64, GS_BASE,    0x12000000);
	GS_Rel(u64, PMODE,    0x0000);	//unwanted trap
	GS_Rel(u64, SMODE1,   0x0010);	//not used
	GS_Rel(u64, SMODE2,   0x0020);	//patch trap
	GS_Rel(u64, SRFSH,    0x0030);	//not used
	GS_Rel(u64, SYNCH1,   0x0040);	//unwanted trap
	GS_Rel(u64, SYNCH2,   0x0050);	//not used
	GS_Rel(u64, SYNCV,    0x0060);	//patch trap
	GS_Rel(u64, DISPFB1,  0x0070);	//not used
	GS_Rel(u64, DISPLAY1, 0x0080);	//patch trap
	GS_Rel(u64, DISPFB2,  0x0090);	//not used
	GS_Rel(u64, DISPLAY2, 0x00A0);	//patch trap
	GS_Rel(u64, EXTBUF,   0x00B0);	//not used
	GS_Rel(u64, EXTDATA,  0x00C0);	//unwanted trap
	GS_Rel(u64, EXTRITE,  0x00D0);	//not used
	GS_Rel(u64, BGCOLOR,  0x00E0);	//unwanted trap
//-----
	GS_Rel(u64, CSR,      0x0400);	//not used
	GS_Rel(u64, IMR,      0x0410);	//not used
//-----
	GS_Rel(u64, BUSDIR,   0x0440);	//not used
//-----
	GS_Rel(u64, SIGLBLID, 0x0480);	//not used
//-----
	GS_Rel(u64, SYSCNT,   0x04F0);	//not used
//-----

// TSR - Code Section
// ------------------

void *KSEG_next_func_p = &KSEG_func_base;

//----------------------------------------------------------------------------
//Here start the function segments intended for KSEG
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Start of ModdedSetGsCrt stuff
//----------------------------------------------------------------------------
KFun_start(ModdedSetGsCrt)
	".word	0\n"
	".word	0\n"
	".word	XBRA_MAGIC\n"
	".word	GSM__MAGIC\n"
	".word	0\n"
KFun_entry(ModdedSetGsCrt)
	"la $v0, KSeg\n"	// Upper Immediate 8 = 0x80000 = Base Address
	"addiu $sp, $sp, -0x0010\n"		// reserve 16 bytes stack space (1 reg)
	"sw		$a0, rel_Source_Interlace($v0)\n"
	"sw		$a1, rel_Source_Mode($v0)\n"
	"sw		$a2, rel_Source_Field($v0)\n"
	"and	$a3, $a0,$a2\n"			//a3 = Interlace & Field
	"andi $a3, $a3,1\n"				//a3 &= 1 limited to 1 bit
	"sb		$a3, rel_Adapt_DoubleHeight($v0)\n"
	"ld $a0, rel_Target_SMode2($v0)\n"
	"sb		$a0, rel_Adapt_SMode2($v0)\n"
	"lw		$a0, rel_Target_Interlace($v0)\n"
	"lw		$a1, rel_Target_Mode($v0)\n"
	"lw		$a2, rel_Target_Field($v0)\n"
	"lw		$a3, rel_sceSetGsCrt($v0)\n"		// a3 -> original SetGsCrt function
	"sd $ra, 0($sp)\n"	// Push return address on stack
	"jalr $a3\n"				// Call original SetGsCrt
	"nop\n" 
	"ld $ra, 0($sp)\n"	// Pull return address from stack 
	"jr		$ra\n"					// Return to caller
	"addiu $sp, $sp, 0x0010\n"	// Restore sp during return
KFun_end(ModdedSetGsCrt);
//----------------------------------------------------------------------------
// End of ModdedSetGsCrt stuff
//----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Start of DisplayHandler stuff
//----------------------------------------------------------------------------
/*
When the processor takes a level 2 exception, the processor switches to
kernel mode, by setting Status.ERL to 1.
*/
KFun_start(DisplayHandler)
	".word	0\n"
	".word	0\n"
	".word	XBRA_MAGIC\n"
	".word	GSM__MAGIC\n"
	".word	0\n"
KFun_entry(DisplayHandler)
	".set noat\n"
	".set nomacro\n"
	"sync.l\n"
	"sync.p\n"
	"sq $k0, -0x10($zero)\n"		// Store registers reserved for kernel
	"sq $k1, -0x20($zero)\n"		// usage in interrupt/trap handling 
	//Save all MIPS registers except zero($0) k0($26) and k1($27))
	//RA NB: NO!!! ALL registers are needed in this array, for evaluations
	//RA NB: Even the $zero register is needed, as it may be used in conditionals
	"la $k0, MIPS_Regs\n"
	"sq $zero, 0($k0)\n"		//$zero
	"sq	$1, 0x10($k0)\n"		//at
	"sq	$2, 0x20($k0)\n"		//v0
	"sq	$3, 0x30($k0)\n"		//v1
	"sq	$4, 0x40($k0)\n"		//a0
	"sq	$5, 0x50($k0)\n"		//a1
	"sq	$6, 0x60($k0)\n"		//a2
	"sq	$7,	0x70($k0)\n"		//a3
	"sq	$8, 0x80($k0)\n"
	"sq	$9, 0x90($k0)\n"
	"sq	$10, 0xA0($k0)\n"
	"sq	$11, 0xB0($k0)\n"
	"sq	$12, 0xC0($k0)\n"
	"sq	$13, 0xD0($k0)\n"
	"sq	$14, 0xE0($k0)\n"
	"sq	$15, 0xF0($k0)\n"
	"sq	$16, 0x100($k0)\n"
	"sq	$17, 0x110($k0)\n"
	"sq	$18, 0x120($k0)\n"
	"sq	$19, 0x130($k0)\n"
	"sq	$20, 0x140($k0)\n"
	"sq	$21, 0x150($k0)\n"
	"sq	$22, 0x160($k0)\n"
	"sq	$23, 0x170($k0)\n"
	"sq	$24, 0x180($k0)\n"
	"sq	$25, 0x190($k0)\n"
	//0x1A0 must be set later, to initial $k0 value
	//0x1B0 must be set later, to initial $k1 value
	"sq	$28, 0x1C0($k0)\n"
	"sq	$29, 0x1D0($k0)\n"
	"sq	$30, 0x1E0($k0)\n"
	"sq	$31, 0x1F0($k0)\n"

	"lq $t0, -0x10($zero)\n"		//t0 = entry k0
	"lq	$t1, -0x20($zero)\n"		//t1 = entry k1
	"sq	$t0, 0x1A0($k0)\n"			//store entry k0 in register array
	"sq	$t1, 0x1B0($k0)\n"			//store entry k1 in register array
	
	"beqzl	$zero,main_DisplayHandler\n"	//Jump past the data block
	"nop\n"

//NB: This table needs to be here, because idiot GCC hates some forward refs.
//    It's like an opcode emulation jump table, to differentiate between
//    different kinds of access that may have triggered the debug trap we use.
//    This way the number of cases does not affect the time delay for testing.

"op_t:\n"		//First we have a table with jump offsets for opcode dependency
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//00-03
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//04-07
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//08-0B
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//0C-0F
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//10-13
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//14-17
	".word ignore-op_t,ignore-op_t, ldl_op-op_t,ldr_op-op_t\n"	//18-1B
	".word ignore-op_t,ignore-op_t, lq_op-op_t, sq_op-op_t\n"		//1C-1F
	".word lb_op-op_t, lh_op-op_t,  lwl_op-op_t,lw_op-op_t\n"		//20-23
	".word lbu_op-op_t,lhu_op-op_t, lwr_op-op_t,lwu_op-op_t\n"	//24-27
	".word sb_op-op_t, sh_op-op_t,  swl_op-op_t,sw_op-op_t\n"		//28-2B
	".word sdl_op-op_t,sdr_op-op_t, swr_op-op_t,ignore-op_t\n"	//2C-2F
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//30-33
	".word ignore-op_t,ignore-op_t, ignore-op_t,ld_op-op_t\n"		//34-37
	".word ignore-op_t,ignore-op_t, ignore-op_t,ignore-op_t\n"	//38-3B
	".word ignore-op_t,ignore-op_t, ignore-op_t,sd_op-op_t\n"		//3C-3F

"BD_t1:\n"		//Table 1 for branch opcodes when trapping branch delay slot
	".word B_com0-op_t,	B_com1-op_t,	B_J-op_t,			B_JAL-op_t\n"		//00-03
	".word B_BEQ-op_t,	B_BNE-op_t, 	B_BLEZ-op_t,	B_BGTZ-op_t\n"	//04-07
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//08-0B
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//0C-0F
	".word B_BC0x-op_t,	B_BC1x-op_t,	B_skip-op_t,	B_skip-op_t\n"	//10-13
	".word B_BEQL-op_t,	B_BNEL-op_t,	B_BLEZL-op_t,	B_BGTZL-op_t\n"	//14-17
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//18-1B
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//1C-1F
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//20-23
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//24-27
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//28-2B
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//2C-2F
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//30-33
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//34-37
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//38-3B
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//3C-3F

"BD_t2:\n"		//Table 2 for branch sub-opcodes when trapping branch delay slot
	".word B_BLTZ-op_t,	B_BGEZ-op_t,	B_BLTZL-op_t,	B_BGEZL-op_t\n"	//00-03
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//04-07
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//08-0B
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//0C-0F
	".word B_BLTZAL-op_t,B_BGEZAL-op_t,B_BLTZALL-op_t,B_BGEZALL-op_t\n"	//10-13
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//14-17
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//18-1B
	".word B_skip-op_t,	B_skip-op_t,	B_skip-op_t,	B_skip-op_t\n"	//1C-1F

"main_DisplayHandler:\n"
/*
The read/write ErrorEPC register holds the virtual address at which instruction
processing can resume after servicing an error. This address can be:
- The virtual address of the instruction that caused the exception
- The virtual address of the immediately preceding branch or jump instruction
(when the instruction is in a branch delay slot, and the BD2 bit in the Cause
register is set).
*/
	"mfc0 $k1, $13\n"		//k1 = Cause bits of last exception (COP0 reg 13)
	"srl $k1, 30\n"		//k1 is aligned for BD2 (Flags branch delay slot used)
													//1 -> delay slot, 0 -> normal
	"andi $k1, 1\n"			//k1 = BD2
	"sll $k1, 2\n"			//k1 = BD2*4
	"mfc0 $k0, $30\n"		//k0 = ErrorPC (COP0 reg 30) -> MIPS instruction
	"addu $k0, $k1\n"		//Add 4 to opcode address for Branch Delay Slot
										//Next get rt (target register) and write address
										//but first check that the instruction is one we patch
	"lw $v0, 0($k0)\n"			//v0 = MIPS instruction that caused trap

	"srl $v1,$v0,26\n"				//v1 = opcode (range 0x00-0x3F)
	"andi $v1,$v1, 0x003F\n"	//v1 = pure opcode number
	"sll $v1,$v1,2\n"				//v1 = op_num*2 (word offset for jump table)
	"la $a2, KSeg\n"	//a2 -> KSeg
	"lw $a3,rel_DisplayHandler_p($a2)\n"		//a3 -> beg_DisplayHandler
	"addi $a3,$a3,op_t-beg_DisplayHandler\n"	//a3 -> op_t
	"addu $a0,$v1,$a3\n"			//a0 -> active entry in op_t
	"lw $a1,0($a0)\n"				//a1 = offset from op_t to opcode handler
	"addu $a1,$a1,$a3\n"      //a1 -> opcode handler
	"jr $a1\n"							//jump to separate opcode handlers
	"nop\n"										//with v0=instruction, a2->KSeg, a3->op_t

//For the present we ignore read operations (should never happen. Not trapped)
"ldl_op:\n"
"ldr_op:\n"
"lq_op:\n"
"lb_op:\n"
"lh_op:\n"
"lwl_op:\n"
"lw_op:\n"
"lbu_op:\n"
"lhu_op:\n"
"lwr_op:\n"
"lwu_op:\n"
"lq_op:\n"
"ld_op:\n"
"ignore:\n"		//We just ignore weird opcodes that we don't implement
	"beqzl $zero,exit_DisplayHandler\n"
	"nop\n"

//For the present we treat all write operations as 'sd'
"sq_op:\n"
"sb_op:\n"
"sh_op:\n"
"swl_op:\n"
"sw_op:\n"
"sdl_op:\n"
"sdr_op:\n"
"swr_op:\n"
"sd_op:\n"
"have_some_write:\n"				//Opcode is a write, so we must check further
	"srl $a1, $v0, 16\n"			
	"andi $a1, $a1, 0x1f\n"		//a1 = unscaled rt reg index
	"srl $a0, $v0, 21\n"
	"andi $a0, $a0, 0x1f\n"		//a0 = unscaled base reg index

	"sll $k0, $a0, 4\n"			//k0 = raw base_ix << 4 (scaled base_ix reg index)

	"addu $v1, $a2, $k0\n"		//v1 = &MIPS_Regs[base_ix]; (if type = u128)
	"lw		$a3, 0($v1)\n"			//a3 = base register value
	"andi $k1, $v0, 0xFFFF\n"	//k1 = offset field of instruction
	"addu $a3, $a3, $k1\n"		//a3 = address which triggered breakpoint

	"sll $k0, $a1, 4\n"			//k0 = raw rt_ix << 4 (scaled rt_ix reg index)
	"addu $v0, $a2, $k0\n"		//v0 = &MIPS_Regs[rt_ix];
	"ld $a1, 0($v0)\n"			//a1 = value in rt

//NB: The trapping method forces us to trap some GS registers we don't want.
//    It is crucial that the writing of those registers proceeds undisturbed.
//    This is handled by the final test case below, at label "not_wanted_reg".

//Here a1=source_data, a2->KSeg, a3=dest_address
	"andi $v0,$a3,0xFFFF\n"							//v0 = dest offset from GS_BASE
	"addi $v1,$v0,-GS_rel_SMODE2\n"
	"beqzl $v1,have_SMODE2_write\n"			//if dest == GS_reg_SMODE2
	"nop\n"
	"addi $v1,$v0,-GS_rel_DISPLAY2\n"
	"beqzl $v1,have_DISPLAYx_write\n"		//if dest == GS_reg_DISPLAY2
	"nop\n"
	"addi $v1,$v0,-GS_rel_DISPLAY1\n"
	"beqzl $v1,have_DISPLAYx_write\n"		//if dest == GS_reg_DISPLAY1
	"nop\n"
	"addi $v1,$v0,-GS_rel_SYNCV\n"
	"beqzl $v1,have_SYNCV_write\n"			//if dest == GS_reg_SYNCV
	"nop\n"
"not_wanted_reg:\n"			//Register unwanted, so perform op unchanged
	"sd $a1,0($a3)\n"		//Store source data unchanged to destination
	"beqzl $zero,exit_DisplayHandler\n"
	"nop\n"
//----------------------------	
/*SMODE2
.----.---.---------.-----------------------------------.
|Name|Pos|Format   |Contents                           |
+----+---+---------+-----------------------------------|
|INT | 0 |int 0:1:0|Interlace Mode Setting             |
|    |   |         |0 Non-Interlace Mode               |
|    |   |         |1 Interlace Mode                   |
|FFMD| 1 |int 0:1:0|Setting in Interlace Mode          |
|    |   |         |0 FIELD Mode(Read every other line)|
|    |   |         |1 FRAME Mode(Read every line)      |
|DPMS|3:2|int 0:2:0|VESA DPMS Mode Setting             |
|    |   |         |00 On          10 Suspend          |
|    |   |         |01 Stand-by    11 Off              |
^----^---^---------^-----------------------------------.*/

"have_SMODE2_write:\n"
	"sd $a1,rel_Source_SMode2($a2)\n"		//Source_SMode2 = a1
	"lb $v0,rel_Adapt_Flag_2($a2)\n"		//v0 = Adapt_Flag_2
	"bnel $v0,$zero,store_v0_as_SMODE2\n"	//if Separate SMODE2 fix disabled
	"or		$v0,$zero,$a1\n"								//	go use v0=a1 for SMODE2
	"srl $v0,$a1,1\n"									//v0 = a1 aligned for FFMD in bit0
	"and	$v0,$v0,$a1\n"								//v0 bit 0 = INT & FFMD
	"andi $v0,$v0,1\n"									//v0 bit 0 = INT & FFMD isolated
	"sb		$v0,rel_Adapt_DoubleHeight($a2)\n"	//store Adapt_DoubleHeight flag
	"beqz	$v0,1f\n"											//if no DoubleHeight need
	"ld $v0,rel_Target_SMode2($a2)\n"	//go use Target_SMode2 as adapted SMode2
																				//else just set v0 = Target_SMode2
 	"andi $a1,$a1,2\n"									//a1 = FFMD of Source_SMode2
 	"andi $v0,$v0,0xFFFD\n"							//v0 = Target_SMode2 without FFMD
 	"or		$v0,$v0,$a1\n"								//v0 = Target_SMode2 + Source FFMD
"1:\n" //Here v0 is adapted SMode2 value
	"sb		$v0,rel_Adapt_SMode2($a2)\n"			//Remember this adaption for later
"store_v0_as_SMODE2:\n"
	"beqzl	$zero,exit_DisplayHandler\n"		//Now go exit
 	"sd $v0,0($a3)\n"											//after storing GS_reg_SMODE2

//----------------------------
/*SYNCV
.----.-----.----------.
|Name|Pos. |Format    |
|----+-----+----------+
|VFP | 9:0 |int 0:10:0|
|VFPE|19:10|int 0:10:0|
|VBP |31:20|int 0:12:0|
|VBPE|41:32|int 0:12:0|
|VDP |52:42|int 0:11:0|
|VS  |??:53|int 0:??:0|
'----^-----^----------^*/

"have_SYNCV_write:\n"
	"sd $a1,rel_Source_SyncV($a2)\n"			//Source_SyncV = a1
	"lb $v0,rel_Adapt_Flag_3($a2)\n"		//v0 = Adapt_Flag_3
	"bnel $v0,$zero,store_v0_as_SYNCV\n"	//if Separate SYNCV fix disabled
	"or		$v0,$zero,$a1\n"								//	go use v0=a1 for SYNCV
	"ld $v0, rel_Target_SyncV($a2)\n"			//v0 = Target_SyncV
	"beql $v0,$zero,exit_DisplayHandler\n"	//if Target_SyncV is zero
 	"sd $a1,0($a3)\n"											//	go use Source_SyncV
"store_v0_as_SYNCV:\n"
	"beqzl	$zero,exit_DisplayHandler\n"		//Now go exit
 	"sd $v0,0($a3)\n"											//after storing GS_SYNCV

//----------------------------
"have_DISPLAYx_write:\n"	//Here a1=source_data, a2->KSeg, a3=dest_adress
	"sd $a1, rel_Safe_Orig_Display($a2)\n"	//request DISPLAYx value = a1
	"ld $v1, rel_Target_Display($a2)\n"			//v1=forcing DISPLAYx template

//Safe_Orig_Display == Requested   DX, DY, MAGH, MAGV, DW and DH values
//Target_Display == Modded(forced) DX, DY, MAGH, MAGV, DW and DH values
//Both are 64 bit units with encoded bit fields like GS DISPLAYx registers

//Patch to adapt request to enforced mode in v1 MUST preserve a1, a2, a3

	"lb $v0,rel_Adapt_Flag_0($a2)\n"
	"bnel $v0,$zero,91f\n"						//if(Adapt_Flag_0)
	"or $a1,$zero,$v1\n"      				//	simulate request same as forced mode
"91:\n"

	"li	$v0,0\n"						// preclear v0 as result DISPLAYx accumulator

//Here a0=free, a1=Source_Display, a2->KSeg, a3=dest_address
//Also v0=result_accumulator, v1=Target_Display, t0-t7=free

#include	"Adapt_X.c"
#include	"Adapt_Y.c"

"Adapt_Calced:\n"
	"sd $v0,rel_Calc_Display($a2)\n"  //Store new DISPLAYx value (for feedback)

//End of Patch to adapt request with the resulting request in v0

	"lb $v1,rel_Adapt_Flag_1($a2)\n"		//v1 = Adapt_Flag_1
	"bnel $v1,$zero,94f\n"								//if(Adapt_Flag_1)
	"ld $v0,rel_Target_Display($a2)\n"	//	use forced mode without adaption
"94:\n"

	"lui $a0,(GS_BASE >> 16)\n"				//a0 -> GS_BASE
	"lb $t0,rel_Adapt_SMode2($a2)\n"	//t0 = adapted SMODE2 value
	"sd $t0,GS_rel_SMODE2($a0)\n"			//store it in GS_reg_SMODE2

//	"sd $v0,0($a3)\n"									//Store modified GS_reg_DISPLAYx
	"sd $v0,GS_rel_DISPLAY2($a0)\n"					//Store modified GS_reg_DISPLAY2
	"sd $v0,GS_rel_DISPLAY1($a0)\n"					//Store modified GS_reg_DISPLAY2

	"ld $t0,rel_Target_SyncV($a2)\n"	//t0 = Target_SyncV
	"bnel $t0,$zero,96f\n"							//if Target_SyncV is non-zero
 	"sd $t0,GS_rel_SYNCV($a0)\n"			//	store it in GS_reg_SYNCV
"96:\n"

"exit_DisplayHandler_complex:\n"
//----- Here we restore some registers, used for complex calculations above
	"lq $10, 0xA0($a2)\n"		//t2
	"lq $11, 0xB0($a2)\n"		//t3
	"lq $12, 0xC0($a2)\n"		//t4
	"lq $13, 0xD0($a2)\n"		//t5
	"lq $14, 0xE0($a2)\n"		//t6
	"lq $15, 0xF0($a2)\n"		//t7
"exit_DisplayHandler:\n"
//----- Here we restore most registers used for all DisplayHandler traps
//----- Since only a few registers are used this way we only restore those
	"la $k0, MIPS_Regs\n"	//Restore MIPS_Regs via k0
	"lq $1, 0x10($k0)\n"		//at
	"lq $2, 0x20($k0)\n"		//v0
	"lq $3, 0x30($k0)\n"		//v1
	"lq $4, 0x40($k0)\n"		//a0
	"lq $5, 0x50($k0)\n"		//a1
	"lq $6, 0x60($k0)\n"		//a2
	"lq $7, 0x70($k0)\n"		//a3

//Past this point in DisplayHandler, use only k0,k1,t0,t1

	"mfc0 $k0, $13\n"	//k0 = Cause of last exception
	"srl $k0, 30\n"	//BD2 Flags debug exception in branch delay slot.
							//1 -> delay slot, 0 -> normal
	"andi $k0, 1\n"		//k0 = BD2 bit isolated
	"bnez $k0,	DisplayHandler_BranchDelaySlotException\n"
	// 	Deal properly with Branch Delay Slot Exceptions (when needed)
	"nop\n"

	"mfc0 $k0,$30\n"			//k0 = ErrorEPC
	"addiu $k0,$k0,4\n"		//k0 = ErrorEPC+4 (-> next opcode)
	"mtc0 $k0,$30\n"			//store k0 in Error Exception Program Counter
//	"sync\n"								//ensure COP0 register update before proceeding

	"b DisplayHandler_Final_Exit\n"
	"nop\n"

//----------------------------
"DisplayHandler_BranchDelaySlotException:\n"
	"mfc0 $k0,$30\n"			//k0 = Error Exception Program Counter
	"lw $k0,0($k0)\n"		//k0 = instruction at EPC location (branch or jump)
	"srl $k1,$k0,26\n"		//k1 = aligned for opcode (range 0x00-0x3F)
	"andi $k1,$k1,0x3F\n"	//k1 = pure opcode number
	"sll $k1,$k1,2\n"		//k1 = op_num*2 (offset for jump table)
	"la $t0, KSeg\n"	//t0 -> KSeg
	"lw $t1,rel_DisplayHandler_p($t0)\n"		//t1 -> beg_DisplayHandler
	"addi $t1,$t1,op_t-beg_DisplayHandler\n"	//t1 -> op_t
	"addu $k1,$k1,$t1\n"				//k1 -> op_t+offset
	"addiu $k1,$k1,BD_t1-op_t\n"	//k1 -> table entry in BD_t1
	"lw $t0,0($k1)\n"					//t0 = table entry from BD_t1
	"addu $t0,$t0,$t1\n"				//t0 -> opcode handler
	"jr $t0\n"								//jump to branch/jump opcode handlers
	"nop\n"												//with k0=instruction, t1->op_t

//----------------------------
"B_com1:\n"			//This group contains 8 different branch operations
	"srl $k1,$k0,16\n"		//k1 = aligned for sub_opcode (range 0x00-0x1F)
	"andi $k1,$k1,0x3F\n"	//k1 = pure sub_opcode number
	"sll $k1,$k1,2\n"		//k1 = sub_op_num*2 (offset for jump table)
	"la $t0, KSeg\n"	//t0 -> KSeg
	"lw $t1,rel_DisplayHandler_p($t0)\n"		//t1 -> beg_DisplayHandler
	"addi $t1,$t1,op_t-beg_DisplayHandler\n"	//t1 -> op_t
	"addu $k1,$k1,$t1\n"				//k1 -> op_t+offset
	"addiu $k1,$k1,BD_t2-op_t\n"	//k1 -> table entry in BD_t2
	"lw $t0,0($k1)\n"					//t0 = table entry from BD_t2
	"addu $t0,$t0,$t1\n"				//t0 -> opcode handler
	"jr $t0\n"								//jump to branch/jump opcode handlers
	"nop\n"												//with k0=instruction, t1->op_t

//----------------------------
"B_com0:\n"			//opcode 0x00 includes both JR and JALR
	"li $t0,0xFC1F07FF\n"		//t0 = bitmask for JALR
	"and $k1,$k0, $t0\n"			//k1 = potential JALR instruction
	"li $t0,9\n"							//t0 = JALR test constant
	"beq $k1,$t0,B_JR_JALR\n"	//if JALR identified, go deal with it
	"nop\n"
	"li $t0,0xFC1FFFFF\n"	//t0 = bitmask for JR
	"and $k1,$k0, $t0\n"		//k1 = potential JR instruction
	"li $t0,8\n"						//t0 = JR test constant
	"bne $k1,$t0,B_skip\n"	//if JR not identified, go skip this code 
	"nop\n"
"B_JR_JALR:\n"	//JR or JALR found, so make register indirect jump
	"srl $k1,$k0, 21\n"			//k1 = aligned for JR/JALR rs register number
	"andi $k1,0x1F\n"				//k1 = register number
	"sll $t1,$k1, 4\n"			//t1 = array index for saved register data
	"la $t0, MIPS_Regs\n"		//t0 -> saved register array
	"addu $t0,$t1\n"					//t0 -> register data of JR/JALR rs register
	"lw $t0,0($t0)\n"			//t0 = jump destination address
	"mtc0 $t0,$30\n"					//store t0 in Error Exception Program Counter
//	"sync\n"										//ensure COP0 register update before proceeding
	"b DisplayHandler_Final_Exit\n"
	"nop\n"

//----------------------------
"B_J:\n"
"B_JAL:\n"
	//Here we have a definite jump with absolute address/4 in instruction
	"li $t0, 0x3FFFFFF\n"	//t0 = bitmask for jump destination bits
	"and $k1, $k0, $t0\n"		//k1 = destination bits (== destination/4)
	"sll $t0, $k1, 2\n"			//t0 = jump destination address
	"mtc0 $t0, $30\n"				//store t0 in Error Exception Program Counter
//	"sync\n"									//ensure COP0 register update before proceeding
	"b DisplayHandler_Final_Exit\n"
	"nop\n"

//----------------------------
//'likely' type branches will only trap on delay slot if branch is taken,
//so for those cases we do not need to make any further tests of conditions
"B_likely:\n"
"B_BGEZL:\n"
"B_BGEZALL:\n"
"B_BLTZL:\n"
"B_BLTZALL:\n"
"B_BEQL:\n"
"B_BNEL:\n"
"B_BLEZL:\n"
"B_BGTZL:\n"
"B_taken:\n"
//Here we have a 'branch taken' operation with relative offset/4 in instruction
	"li $t0, 0xFFFF\n"			//t0 = bitmask for branch offset bits
	"and $k1, $k0, $t0\n"		//k1 = branch offset bits (== offset/4)
	"sll $k1, 2\n"					//k1 = branch offset
	"mfc0 $t0, $30\n"				//t0 = Error Exception Program Counter
	"addiu $t0, 4\n"					//t0 = ErrorEPC+4 (-> address after branch op)
	"addu $t0, $k1\n"				//t0 = jump destination address
	"mtc0 $t0, $30\n"				//store t0 in Error Exception Program Counter
//	"sync\n"									//ensure COP0 register update before proceeding
	"b DisplayHandler_Final_Exit\n"
	"nop\n"

//----------------------------
"B_BLTZ:\n"
"B_BLTZAL:\n"
	"srl $k1,$k0, 21\n"		//k1 = aligned for rs register number
	"andi $k1,0x1F\n"				//k1 = register number
	"sll $t1,$k1, 4\n"			//t1 = array index for saved register data
	"la $t0, MIPS_Regs\n"		//t0 -> saved register array
	"addu $t0,$t1\n"				//t0 -> register data of rs register
	"lq $t1,0($t0)\n"			//t1 = rs register data
//	"sync\n"
	"bltz $t1,B_taken\n"
	"nop\n"
"B_not_taken:\n"
	"mfc0 $k0,$30\n"			//k0 = ErrorEPC
	"addiu $k0,$k0,8\n"		//k0 = ErrorEPC+8 pass branch_op and delay_slot
	"mtc0 $k0,$30\n"			//store k0 in Error Exception Program Counter
//	"sync\n"								//ensure COP0 register update before proceeding
	"b DisplayHandler_Final_Exit\n"
	"nop\n"
	
//----------------------------
"B_BGEZ:\n"
"B_BGEZAL:\n"
	"srl $k1,$k0, 21\n"		//k1 = aligned for rs register number
	"andi $k1,0x1F\n"				//k1 = register number
	"sll $t1,$k1, 4\n"			//t1 = array index for saved register data
	"la $t0, MIPS_Regs\n"		//t0 -> saved register array
	"addu $t0,$t1\n"				//t0 -> register data of rs register
	"lq $t1,0($t0)\n"			//t1 = rs register data
//	"sync\n"
	"bgez $t1,B_taken\n"
	"nop\n"
	"b B_not_taken\n"
	"nop\n"
//----------------------------
"B_BLEZ:\n"
	"srl $k1,$k0, 21\n"		//k1 = aligned for rs register number
	"andi $k1,0x1F\n"				//k1 = register number
	"sll $t1,$k1, 4\n"			//t1 = array index for saved register data
	"la $t0, MIPS_Regs\n"		//t0 -> saved register array
	"addu $t0,$t1\n"				//t0 -> register data of rs register
	"lq $t1,0($t0)\n"			//t1 = rs register data
//	"sync\n"
	"blez $t1,B_taken\n"
	"nop\n"
	"b B_not_taken\n"
	"nop\n"
//----------------------------
"B_BGTZ:\n"
	"srl $k1,$k0, 21\n"		//k1 = aligned for rs register number
	"andi $k1,0x1F\n"				//k1 = register number
	"sll $t1,$k1, 4\n"			//t1 = array index for saved register data
	"la $t0, MIPS_Regs\n"		//t0 -> saved register array
	"addu $t0,$t1\n"				//t0 -> register data of rs register
	"lq $t1,0($t0)\n"			//t1 = rs register data
//	"sync\n"
	"bgtz $t1,B_taken\n"
	"nop\n"
	"b B_not_taken\n"
	"nop\n"
//----------------------------
"B_BEQ:\n"
	"srl $k1,$k0, 21\n"		//k1 = aligned for rs register number
	"andi $k1,0x1F\n"				//k1 = rs register number
	"sll $k1,$k1, 4\n"			//k1 = array index for saved rs register data
	"la $t0, MIPS_Regs\n"		//t0 -> saved register array
	"addu $k1,$t0\n"				//k1 -> register data of rs register
	"lq $t1,0($k1)\n"			//t1 = rs register data
//	"sync\n"
	"srl $k1,$k0, 16\n"		//k1 = aligned for rt register number
	"andi $k1,0x1F\n"				//k1 = rt register number
	"sll $k1,$k1, 4\n"			//k1 = array index for saved rt register data
	"addu $k1,$t0\n"				//k1 -> register data of rt register
	"lq $t0,0($k1)\n"			//t0 = rt register data
//	"sync\n"
	"beq $t0,$t1,B_taken\n"
	"nop\n"
	"b B_not_taken\n"
	"nop\n"
//----------------------------
"B_BNE:\n"
	"srl $k1,$k0, 21\n"		//k1 = aligned for rs register number
	"andi $k1,0x1F\n"				//k1 = rs register number
	"sll $k1,$k1, 4\n"			//k1 = array index for saved rs register data
	"la $t0, MIPS_Regs\n"		//t0 -> saved register array
	"addu $k1,$t0\n"				//k1 -> register data of rs register
	"lq $t1,0($k1)\n"			//t1 = rs register data
//	"sync\n"
	"srl $k1,$k0, 16\n"		//k1 = aligned for rt register number
	"andi $k1,0x1F\n"				//k1 = rt register number
	"sll $k1,$k1, 4\n"			//k1 = array index for saved rt register data
	"addu $k1,$t0\n"				//k1 -> register data of rt register
	"lq $t0,0($k1)\n"			//t0 = rt register data
//	"sync\n"
	"bne $t0,$t1,B_taken\n"
	"nop\n"
	"b B_not_taken\n"
	"nop\n"
//----------------------------
"B_BC0x:\n"					//At present we do not implement COP0 branches
"B_BC1x:\n"					//At present we do not implement COP1 branches
"B_skip:\n"					//Unrecognized opcode, so just pass it by
	"mfc0 $k0, $30\n"			//k0 = ErrorEPC
	"addiu $k0, $k0, 4\n"	//k0 = ErrorEPC+4 (-> next opcode)
	"mtc0 $k0, $30\n"			//store k0 in Error Exception Program Counter
//	"sync\n"								//ensure COP0 register update before proceeding

"DisplayHandler_Final_Exit:\n"

	"la $k0, KSeg\n"	//k0 -> KSeg
	"lq $8, 0x80($k0)\n"		//t0
	"lq $9, 0x90($k0)\n"		//t1

//	"mfc0 $k0, $12\n"				// Set user mode, interrupts on
//	"ori $k0, $k0, 0x11\n"
//	"mtc0 $k0, $12\n"
//	"sync\n"

	"lq $k0, -0x10($zero)\n"		// Restore k0,k1 reserved for OS Kernel
	"lq $k1, -0x20($zero)\n"
	"sync.p\n"
	"sync.l\n"
	"eret\n"								// 	Return from exception
	"nop\n"

"end_DisplayHandler:\n"
	".set macro\n"
	".set at\n"
	KFun_end(DisplayHandler);
//----------------------------------------------------------------------------
// End of DisplayHandler stuff
//----------------------------------------------------------------------------

//Here end the function segments intended for KSEG
//----------------------------------------------------------------------------
u64 make_display_magic_number(dh, dw, magv, magh, dy, dx)
{
	u64 rv;

	rv = ((u64)dx      & ((1<<12)-1))
	    |(((u64)dy     & ((1<<11)-1))<<12)
	    |(((u64)magh   & ((1<<4)-1)) <<23)
	    |(((u64)magv   & ((1<<2)-1)) <<27)
	    |((((u64)dw-1) & ((1<<12)-1))<<32)
	    |((((u64)dh-1) & ((1<<11)-1))<<44);

//RA NB: dw and dh are here 1 unit higher than the real GS register values
//RA NB: so for a 480p resolution we use 480, though the register gets 479
//RA NB: which is thus what we here embed in the bits of the return value

	return rv;
}//ends make_display_magic_number
//----------------------------------------------------------------------------
unsigned int get_sr() {
    register unsigned int rv;
	__asm__ __volatile__ (
	".set noreorder\n"
    "mfc0 %0, $12\n"
    "sync\n"
    "nop\n"
	".set reorder\n"
    : "=r" (rv) : );
    return rv;
}

//----------------------------------------------------------------------------
void set_sr(unsigned int v) {
	__asm__ __volatile__ (
	".set noreorder\n"
    "mtc0 %0, $12\n"
    "sync\n"
    "nop\n"
	".set reorder\n"
    : : "r" (v) );
}

//----------------------------------------------------------------------------
void set_kmode() {
	set_sr(get_sr() & (~0x19));
}

//----------------------------------------------------------------------------
void set_umode() {
	set_sr((get_sr() & (~0x19)) | 0x11);
}

//----------------------------------------------------------------------------
int disableClearUserMem()
{
	int *scanMem = (int *)0x80000000;
	int retval = -1;
	int i;
	set_kmode();
	for(i = 0; i < 0x3FFFF; i++)
	{
	 if( ((scanMem[i] & 0xffff0000) == 0x3C040000) &&	 	// lui a0, 0000
		 ((scanMem[i + 1] & 0x7C000000) == 0xC000000) &&	// jal XXXX
		 (scanMem[i + 2] == 0x34842000) )				 	// ori a0, a0, 0x2000
		 {
			 scanMem[i + 1] = NOP;
			 retval = 0;
			 break;
		 }
	}
	set_umode();
	return retval;
}

//----------------------------------------------------------------------------
void delay(int count)
{
	int i;
	int ret;
	for (i = 0; i < count; i++) {
		ret = 0x01000000;
			while(ret--) asm("nop\nnop\nnop\nnop");
	}
}

//----------------------------------------------------------------------------
void poweroffHandler(int i)
{
   hddPowerOff();
}

//----------------------------------------------------------------------------
/* Read IOP modules */
void LoadModules(void) {
	/* Load memory card driver (IOP modules) - WARNING: Do not change this calling order! */
	SifLoadModule("rom0:SIO2MAN", 0, NULL);	// Basic serial communications module
	SifLoadModule("rom0:MCMAN", 0, NULL);	// MC (Memory card) control module #1
	SifLoadModule("rom0:MCSERV", 0, NULL);	// MC module #2
	mcInit(MC_TYPE_MC);						// Initialize MC driver
	SifLoadModule("rom0:PADMAN", 0, NULL);	// Controller control module
}

//----------------------------------------------------------------------------
/* This allow KSEG (Kernel EE RAM Area) Memory Copy */
u32 KSEG_memcpy(void *addr, void *buf, u32 size)
{
   DI();						// Disable interrupts
   set_kmode();					// Enter kernel mode
   memcpy(buf, addr, size);		// Memory Copy
   set_umode();					// Leave kernel mode
   EI();						// Enable interrupts
   return size;
}

//----------------------------------------------------------------------------
/*  Output Memory Dump to Screen - I recognize I have no skill, no time, and no conditions to learn / use naplink
    So I using this piece of trash code to debug the results on screen
    (Always when needed, I can take photos of it using my mobile and change code later) */
void MemoryDump(u32 Start, u32 Quantity)
{
        u32 i;
        u8 Data;

		scr_printf("[%08x] ", Start);		// Display Start Address
        
        for (i = 0; i < Quantity; i++)
        {
        	set_kmode();
        	Data = _lb(Start + i);
        	set_umode();
			scr_printf("%02x ", Data);		// Display Address Data
        }
		scr_printf("\n");					// Line break
}

//----------------------------------------------------------------------------
/* Get Syscall Vector Table Address - The value is different among PS2 versions, this is the reason why we must search for that! */
u32 GetSyscallVectorTableAddress(void)
{
   // Here, we set magical numbers (instead of ordinary address) to system call vector table positions 0xFE and 0xFF 
   // (both are related to syscalls code numbers that are unused). Doing this, i.e., puting non-usual opcodes sequence, 
   // that probably does not exist on original code, let us use this magic numbers (fake addresses) to found
   // Syscall Vector Table Start Address. It is used "BIG BOOBS CODE" here (0x1337C0DE+0xB16B00B5) but it could be
   // anything else that fits on this concept!
   const s32 syscall_num[2] = { 0xFE, 0xFF };
   const u32 magic[2] = { 0x1337C0DE, 0xB16B00B5 };	
   u32 addr = -1;
   u32 i;
   SetSyscall(syscall_num[0], (void*)magic[0]);
   SetSyscall(syscall_num[1], (void*)magic[1]);
   set_kmode();
   // Scan start address
   for (i = 0x80000000; i < 0x80080000; i += 4) {
      if (!memcmp((u32*)i, magic, sizeof(magic))) {
         addr = i - syscall_num[0] * 4;
         break;
      }
   }
   set_umode();
   SetSyscall(syscall_num[0], (void*)0);
   SetSyscall(syscall_num[1], (void*)0);
   return addr;
} 

//----------------------------------------------------------------------------
/* Set Syscall */
void setSyscall(int number, void (*functionPtr)(void)) {
	__asm__ __volatile__ (
	".set noreorder\n"
	".set noat\n"
	"li $3, 0x74\n"
    "add $4, $0, %0    \n"   // Specify the argument #1
    "add $5, $0, %1    \n"   // Specify the argument #2
   	"syscall\n"
	"jr $31\n"
	"nop\n"
	".set at\n"
	".set reorder\n"
    :
    : "r"( number ), "r"( functionPtr )
    );
}

//----------------------------------------------------------------------------
void setNormalAdaption(void) //Should be called before exiting main program
{
	set_kmode();
	Adapt_Flags &= 0xFFFFFF00; //Clear 1 byte flag of 4
  set_umode();

}

//----------------------------------------------------------------------------
//Use XBRA protocol to unlink an old instance of GSM, if any such is active
void unlink_GSM()
{
	u32 SyscallVectorTableAddress;
	u32 SetGsCrtVectorAddress;
	u32 SetGsCrt_Addr;
	XBRA_header *setGsCrt_XBRA_p;

	SyscallVectorTableAddress = GetSyscallVectorTableAddress();
	SetGsCrtVectorAddress = SyscallVectorTableAddress + 2 * 4;
	
	set_kmode();
	syscallTable = (void *) SyscallVectorTableAddress;
	SetGsCrt_Addr = _lw(SetGsCrtVectorAddress);
	setGsCrt_XBRA_p = (XBRA_header *)(SetGsCrt_Addr - XB_code);

	if((setGsCrt_XBRA_p -> xb_magic == XBRA_MAGIC)
	&& (setGsCrt_XBRA_p -> xb_id == GSM__MAGIC)
	){
		//Here we need to abort an old install, as GSM is already active.
		//So we unlink that old install before linking in the new one
		//It has to be done this way because the new instance may execute
		//in a different KSEG location, for game compatibility reasons
		//Note that similar unlinking is not needed for DisplayHandler,
		//since this debug trap routine is not chained to any previous instance,
		//as each new instance completely replaces any old one.
		//But care must still be taken so as not to crash the console by making
		//vmode changes through setGsCrt in a half initialized state
		SetGsCrt_Addr = setGsCrt_XBRA_p -> xb_next;
		_sw(SetGsCrt_Addr, SetGsCrtVectorAddress);	//unlink setGsCrt of old GSM
	}
}

//----------------------------------------------------------------------------
/* Install GS (Graphics Sinthesizer) Mode Selector */
void installGSModeSelector(u32 interlace, u32 mode, u32 field, u64 display, u64 syncv, u64 smode2)
{
	u32 SyscallVectorTableAddress;
	u32 SetGsCrtVectorAddress;
	u32 SetGsCrt_Addr;
	XBRA_header *setGsCrt_XBRA_p;

	SyscallVectorTableAddress = GetSyscallVectorTableAddress();
	SetGsCrtVectorAddress = SyscallVectorTableAddress + 2 * 4;
	
	scr_printf("Syscall Vector Table found at 0x%08x\n", SyscallVectorTableAddress);
	MemoryDump(SyscallVectorTableAddress, 0x0C);
	   
	/* Install sceGsSetCrt patcher */
	/* Syscall hooked: Syscall #2 (SetGsCrt) */
	set_kmode();
	Adapt_Flags = 0x00000001;						//Init 4 byte flags
	Target_Interlace = interlace;
	Target_Mode = mode;
	Target_Field = field;
  set_umode();
  	
	scr_printf("Stored parameters to be forced by ModdedSetGsCrt (Interlace, Mode, Field) at\n");
	MemoryDump((u32) &Target_Interlace, 0x0C);
	
	set_kmode();
	Target_Display = display;
	set_umode();

	scr_printf("Stored parameters to be forced by DisplayHandler (Display) at\n");
	MemoryDump((u32) &Target_Display, 0x08);
	
	set_kmode();
	Target_SyncV = syncv;
	set_umode();

	scr_printf("                                                 (SyncV) at\n");
	MemoryDump((u32) &Target_SyncV, 0x08);

	set_kmode();
	Target_SMode2 = smode2;
	Adapt_SMode2 = (u8) smode2;
	set_umode();

	scr_printf("                                                 (SMode2) at\n");
	MemoryDump((u32) &Target_SMode2, 0x08);

	set_kmode();
	syscallTable = (void *) SyscallVectorTableAddress;
	SetGsCrt_Addr = _lw(SetGsCrtVectorAddress);
	setGsCrt_XBRA_p = (XBRA_header *)(SetGsCrt_Addr - XB_code);

	//Note that this routine is no longer responsible for unlinking old instances
	//It is assumed that this is done previously by a call to unlink_GSM
	//performed before the new routines were copied to KSEG RAM (already done)

	//Here we are ready to link in the new setGsCrt routine of a new GSM instance
	setGsCrt_XBRA_p = (XBRA_header *)(KSEG_ModdedSetGsCrt_entry_p - XB_code);
	sceSetGsCrt = (void *) SetGsCrt_Addr;
	setGsCrt_XBRA_p -> xb_next = (u32) sceSetGsCrt;
	_sw((int)KSEG_ModdedSetGsCrt_entry_p, SetGsCrtVectorAddress);
	//SyscallVectorTableAddress[2] = (int)KSEG_ModdedSetGsCrt_entry_p;

	set_umode();

	scr_printf("[BEFORE]SyscallVectorTableAddress[2] =       SetGsCrt = 0x%08x\n", SetGsCrt_Addr);
	scr_printf("[AFTER ]SyscallVectorTableAddress[2] = ModdedSetGsCrt = 0x%08x\n\n", (int)KSEG_ModdedSetGsCrt_entry_p);
	MemoryDump(SyscallVectorTableAddress, 0x0C);

	scr_printf("[BEFORE] Core Debug Exception Handler (V_DEBUG handler)\n");
	MemoryDump(0x80000100, 0x08);
	/* Install Display Handler in place of the Core Debug Exception Handler (V_DEBUG handler) */
	/* Exception Vector Address for Debug Level 2 Exception when Stadus.DEV bit is 0 (normal): 0x80000100 */
	/* 'Level 2' is a generalization of Error Level (from previous MIPS processors) */
	/* When this exception is recognized, control is transferred to the applicable service routine;
	   in our case the service routine is 'DisplayHandler'! */
	
	set_kmode();
	/* Do you know that '*((unsigned int*)Adddr) = Data' has the same effect that '_sw(Data), Address)'? ;-b */
	//*((unsigned int*)0x80000100) = MAKE_J((int)KSEG_DisplayHandler_entry_p);
	//*((unsigned int*)0x80000104) = 0;
	_sw(MAKE_J((int)KSEG_DisplayHandler_entry_p), 0x80000100);
	_sw(0, 0x80000104);
	//flush_cache();
	set_umode();
	scr_printf("[AFTER ] DisplayHandler -> J 0x%08x\n", (int)KSEG_DisplayHandler_entry_p);
	MemoryDump(0x80000100, 0x08);
    
	/* Set Data Address Write Breakpoint */
	/* Trap writes to GS registers, so as to control their values */
	__asm__ __volatile__ (
	".set noreorder\n"
	".set noat\n"
	
	"li $a0, 0x12000000\n"	// Address
	"li $a1, 0xFFFFFF1F\n"	// Mask
	//We trap writes to 0x12000000 + 0x00,0x20,0x40,0x60,0x80,0xA0,0xC0,0xE0
	//We only want 0x20, 0x60, 0x80, 0xA0, but can't mask for that combination

	"di\n"									// Disable Interupts
	"sync\n"								// Wait until the preceding loads are completed

	"ori $k0, $zero, 0x8000\n"
	"mtc0 $k0, $24\n"			// All breakpoints off (BED = 1)
	"sync.p\n"						// Await instruction completion
	"mtdab	$a0\n"
	"mtdabm	$a1\n"
	"sync.p\n"						// Await instruction completion
	"mfc0 $k1, $24\n"
	"sync.p\n"						// Await instruction completion

	"lui $k0, 0x2020\n"			// Data write breakpoint on (DWE, DUE = 1)
	"or $k1, $k1, $k0\n"
	"xori $k1, $k1, 0x8000\n"		// DEBUG exception trigger on (BED = 0)
	"mtc0 $k1, $24\n"
	"sync.p\n"						//  Await instruction completion

	"ei\n"						// Enable Interupts
	"nop\n"
	"nop\n"
	
	".set at\n"
	".set reorder\n"
	);
	scr_printf("Now, writes to the GS registers we need to control are trapped!\n\n");

}

//----------------------------------------------------------------------------
/* Update parameters to be forced by ModdedSetGsCrt and DisplayHandler TSR functions */
void ModdedSetGsCrtDisplayHandlerParams(u32 interlace, u32 mode, u32 field, u64 display, u64 syncv, u64 smode2)
{
	set_kmode();
	Target_Interlace = interlace;
	Target_Mode = mode;
	Target_Field = field;
	Target_Display = display;
	Target_SyncV = syncv;
	Target_SMode2 = smode2;
	Adapt_SMode2 = (u8) smode2;
  set_umode();
}
//----------------------------------------------------------------------------
void	CleanUp(void)
{
	setNormalAdaption();
	TimerEnd();
	padPortClose(1,0);
	padPortClose(0,0);
	padEnd();
}
//----------------------------------------------------------------------------
int Create_File (char *filepath, u8 *buf, int size) // Creates a file on mc0:
{
	int fd;
	char filepath2[0x40];

	sprintf(filepath2, "mc0:%s", filepath);
	fd = fioOpen(filepath2, O_WRONLY | O_CREAT);
	if(fd < 0)
	{	filepath2[2] = '1'; //change path to mc1:
		fd = fioOpen(filepath2, O_WRONLY | O_CREAT);
		if(fd < 0)
			return 0;
		MC_index = 1;
	} else {
		MC_index = 0;
	}
	if (fioWrite(fd, buf, size) < 0) return 0;
	fioClose(fd);
	
	return 1;
}
//----------------------------------------------------------------------------
int File_Exist (char *filepath) // Check that a file is existing
{
	int fd;
	char filepath2[0x40];

	sprintf(filepath2, "mc0:%s", filepath);
	fd = fioOpen(filepath2, O_RDONLY);
	if(fd < 0)
	{	filepath2[2] = '1'; //change path to mc1:
		fd = fioOpen(filepath2, O_RDONLY);
		if(fd < 0)
			return 0;
		MC_index = 1;
	} else {
		MC_index = 0;
	}
	fioClose(fd);	
	
	return 1;
}
//----------------------------------------------------------------------------
//________________ From uLaunchELF ______________________
//---------------------------------------------------------------------------
// get_CNF_string is the main CNF parser called for each CNF variable in a
// CNF file. Input and output data is handled via its pointer parameters.
// The return value flags 'false' when no variable is found. (normal at EOF)
//---------------------------------------------------------------------------
int	get_CNF_string(unsigned char **CNF_p_p,
                   unsigned char **name_p_p,
                   unsigned char **value_p_p)
{
	unsigned char *np, *vp, *tp = *CNF_p_p;

start_line:
	while((*tp<=' ') && (*tp>'\0')) tp+=1;  	//Skip leading whitespace, if any
	if(*tp=='\0') return 0;            			//but exit at EOF
	np = tp;                               		//Current pos is potential name
	if(*tp<'A')                            		//but may be a comment line
	{                                      		//We must skip a comment line
		while((*tp!='\r')&&(*tp!='\n')&&(*tp>'\0')) tp+=1;  //Seek line end
		goto start_line;                     	//Go back to try next line
	}

	while((*tp>='A')||((*tp>='0')&&(*tp<='9'))) tp+=1;  //Seek name end
	if(*tp=='\0') return 0;            			//but exit at EOF

	while((*tp<=' ') && (*tp>'\0'))
		*tp++ = '\0';                        	//zero&skip post-name whitespace
	if(*tp!='=') return 0;  	           		//exit (syntax error) if '=' missing
	*tp++ = '\0';                          		//zero '=' (possibly terminating name)

	tp += 1;									//skip '='
	while((*tp<=' ') && (*tp>'\0')				//Skip pre-value whitespace, if any
		&& (*tp!='\r') && (*tp!='\n')			//but do not pass EOL
		&& (*tp!='\7')     						//allow ctrl-G (BEL) in value
		)tp+=1;									//skip tested whitespace characters
	if(*tp=='\0') return 0;						//but exit at EOF
	vp = tp;									//Current pos is potential value	

	while((*tp!='\r')&&(*tp!='\n')&&(*tp!='\0')) tp+=1;  //Seek line end
	if(*tp!='\0') *tp++ = '\0';            		//terminate value (passing if not EOF)
	while((*tp<=' ') && (*tp>'\0')) tp+=1;  	//Skip following whitespace, if any

	*CNF_p_p = tp;                         		//return new CNF file position
	*name_p_p = np;                        		//return found variable name
	*value_p_p = vp;                       		//return found variable value
	return 1;                           		//return control to caller
}	//Ends get_CNF_string
//----------------------------------------------------------------
// Load CNF
//----------------------------------------------------------------
int loadConfig(char *filepath)
{
	int fd, var_cnt, user_index=0;

	size_t CNF_size;
	char path[0x40];
	unsigned char *RAM_p, *CNF_p, *name, *value;
	
	sprintf(path, "mc0:%s", filepath);
	fd = -1;
	fd = fioOpen(path, O_RDONLY);  // Try to open cnf from mc0:
	if (fd < 0) {
		path[2] = '1'; //change path to mc1:
		fd = fioOpen(path, O_RDONLY);  // Try to open cnf from mc1:
		if (fd < 0) {
			return 0;
		}
		MC_index = 1;
	} else {
		MC_index = 0;
	}
	// This point is only reached after succefully opening CNF	 

	CNF_size = lseek(fd, 0, SEEK_END);
	fioLseek(fd, 0, SEEK_SET);
	RAM_p = (char*)malloc(CNF_size);
	CNF_p = RAM_p;
	if (CNF_p == NULL) {
		close(fd);
failed_load:
		return 0;
	}
	fioRead(fd, CNF_p, CNF_size); // Read CNF as one long string
	fioClose(fd);
	CNF_p[CNF_size] = '\0'; // Terminate the CNF string

	for (var_cnt = 0; get_CNF_string(&CNF_p, &name, &value); var_cnt++) {
		//A variable was found, now we dispose of its value.
		//The first variable should always be GSM_CNF_version (defines file content)
		if(var_cnt == 0){ //if this is the first variable found ?
			if (!strcmp(name,"GSM_CNF_version")) { //if its name is GSM_CNF_version?
				strncpy(GSM->GSM_CNF_version, value, 63); //Copy body of value string
				GSM->GSM_CNF_version[63] = 0; //Force value string termination 
				//This value can be used to modify CNF interpretation below
			}else{ //if first variable is NOT GSM_CNF_version, the file is invalid
				goto failed_load;
			}
		}
		if(strcmp(GSM->GSM_CNF_version, CNF_VERSION) != 0)
			goto failed_load;

		//The two variables below are just examples for string VS numeric usage
		if (!strcmp(name,"Option_1")) {
			strncpy(GSM->option_1, value, 63); //Copy body of value string
			GSM->option_1[63] = 0; //Force value string termination 
			continue;
		}		
		if (!strcmp(name,"Option_2")) {
			GSM->option_2 = atoi(value);	//This syntax is used for numeric values
			continue;
		}
		if (!strcmp(name,"User_Index")) {
			user_index = atoi(value) & (user_vmode_slots-1); //Load and limit index
			continue;
		}
		if (!strcmp(name,"interlace")) {
			user_vmode[user_index].interlace = atoi(value);
			continue;
		}
		if (!strcmp(name,"mode")) {
			user_vmode[user_index].mode = atoi(value);
			continue;
		}
		if (!strcmp(name,"field")) {
			user_vmode[user_index].field = atoi(value);
			continue;
		}
		if (!strcmp(name,"display")) {
			user_vmode[user_index].display = atol(value);
			continue;
		}
		if (!strcmp(name,"syncv")) {
			user_vmode[user_index].syncv = atol(value);
			continue;
		}
		if (!strcmp(name,"smode2")) {
			user_vmode[user_index].smode2 = atol(value);
			continue;
		}
	} // ends for
	free(RAM_p);
	//NB: Because the RAM buffer is released here, all string values must be
	//    copied to permanent variable space in the CNF interpretation above.

	return 1;
}
//------------------------------
//endfunc loadConfig
//----------------------------------------------------------------
// Save CNF
//----------------------------------------------------------------
int saveConfig(char *filepath)
{
	int i, ret, fd;
	char tmp[26*MAX_PATH + 30*MAX_PATH];
	char path[0x40];
	size_t CNF_size, CNF_step;

	sprintf(path, "mc0:%s", filepath);
	if(MC_index == 1)
		path[2] = '1';

	CNF_size = 0;
	
	sprintf(tmp+CNF_size,
		"GSM_CNF_version = %s\r\n"
		"Option_1 = %s\r\n"
		"Option_2 = %d\r\n"
		"%n",           // %n causes NO output, but only a measurement
		GSM->GSM_CNF_version,
		GSM->option_1,
		GSM->option_2,
		&CNF_step       // This variable measures the size of sprintf data
  );
	CNF_size += CNF_step;

	for(i=0; i<user_vmode_slots; i++){	//Loop to save the user-defined vmodes
		sprintf(tmp+CNF_size,
			"User_Index = %d\r\n"
			"interlace = %d \r\n"
			"mode = %d \r\n"
			"field = %d\r\n"
			"display = %ld\r\n"
			"syncv = %ld\r\n"
			"smode2 = %ld\r\n"
			"%n",           // %n causes NO output, but only a measurement
			i,
			user_vmode[i].interlace,
			user_vmode[i].mode,
			user_vmode[i].field,
			user_vmode[i].display,
			user_vmode[i].syncv,
			user_vmode[i].smode2,
			&CNF_step       // This variable measures the size of sprintf data
	  );
		CNF_size += CNF_step;
	}//ends for

	fd = fioOpen(path,O_CREAT|O_WRONLY|O_TRUNC);  // Try to open cnf on one MC
	if (fd < 0) {
		path[2] ^= 1; //change path to other MC
		fd = fioOpen(path, O_CREAT|O_WRONLY|O_TRUNC);  // Try to open cnf on other MC
		if (fd < 0) {
			return 0;
		}
	}
	// This point is only reached after succefully opening CNF	 
	MC_index = path[2]-'0';

	ret = fioWrite(fd,&tmp,CNF_size);
	fioClose(fd);
	if(ret==CNF_size)
		return 1;
	else
		return 0;
}
//------------------------------
//endfunc saveConfig
//---------------------------------------------------------------------------
void Set_Default_Settings(void)
{
	strcpy(GSM->GSM_CNF_version, CNF_VERSION);
	strcpy(GSM->option_1, "Default");
	GSM->option_2 = 0;  
}
//---------------------------------------------------------------------------
int detect_PAL_mode() //returns 1 for PAL, but 0 for NTSC
{
	int fd;

	fd = open("rom0:ROMVER", O_RDONLY);
	if(fd < 0)
		return 0; //NTSC is default mode for unidentified console
	read(fd, ROMVER_data, 14); ROMVER_data[14]=0;
	close(fd);

	if(ROMVER_data[4] == 'E')
		return 1; //PAL mode is identified by 'E' for Europe
	return 0;  //All other cases need NTSC
}
//------------------------------
//endfunc detect_PAL_mode
//----------------------------------------------------------------------------
/*
+------------------------------+
|                              |
|   M A I N   F U N C T I O N  |
|                              |
+------------------------------+
*/
int main()
{   

	char elf_path[0x40];
	char party[1] = {0};
   
	u32 size;
	
	u32 interlace, act_interlace;
	u32 mode, act_mode;
	u32 field, act_field;
	u64 display;
	u64 syncv, act_syncv;
	u64 smode2;
	
	char *on_off_s[] = {"ON", "OFF"};

	u64 presets, dispreq, dispcalc;
	
	u32 gs_dx, gs_dy, gs_magh, gs_magv, gs_dw, gs_dh;
//RA NB: gs_dw and gs_dh are 1 unit higher than the real GS register values
//RA NB: so for a 480p resolution we use 480, though the register gets 479

	int updateflag, adapt1, adapt2, adapt3, user_index=0;
	int local_CSR_copy;
	int joy_step, retval=0;

	/* PS2 Configuration settings
			note: the 'set' methods are only valid till the ps2 gets
			turned off or reset! */
	   
	GSM = malloc(sizeof(struct gsm_settings));

	// Set GSM default settings for configureable items
  Set_Default_Settings();

	/* Initialise */
   
	init_scr();
	
	scr_printf("%s - by %s\n", VERSION, AUTHORS);
	scr_printf("--------------------------------------------------------------------\n\n");
		
	scr_printf("Loading access modules...\n");
	LoadModules();

 	scr_printf("Initialising the joypad subsystem...\n");
	TimerInit();
	Timer();
	setupPad();
	TimerInit();
	Timer();

	unlink_GSM();
	KFun_install(ModdedSetGsCrt);
	KFun_install(DisplayHandler);
	set_kmode();
	DisplayHandler_p = KSEG_DisplayHandler_p;
	set_umode();

 	scr_printf("Disabling InitializeUserMemory function call in EE Kernel...\n");
	disableClearUserMem();

	if(detect_PAL_mode()) { //If PAL mode detected (by ROMVER check)
		mode = 0x03;					//	set PAL mode
		interlace = 0x01;			// Interlaced
		field = 0x00;					// Field mode (Read Every Other Line)
		gs_dx=720; gs_dy=68; gs_magh=3; gs_magv=0;
		gs_dw = 2560; gs_dh = 512;
		syncv = 0;
		smode2 = 1;
	} else {								//else
		mode = 0x02;					//	set NTSC mode
		interlace = 0x01;			// Interlaced
		field = 0x00;					// Field mode (Read Every Other Line)
		gs_dx=700; gs_dy=46; gs_magh=3; gs_magv=0;
		gs_dw = 2560; gs_dh = 448;
		syncv = 0;
		smode2 = 1;
	}
	act_interlace = interlace;
	act_mode = mode;
	act_field = field;
	act_syncv = syncv;


	display = make_display_magic_number(gs_dh, gs_dw, gs_magv, gs_magh, gs_dy, gs_dx);

	syncv = 0;		//Disable SYNCV register patching by default
	smode2 = 1;		//set SMODE2 for interlace by default

	/* Install patcher function for sceGsSetCrt and DISPLAY 1/2 GS Registers */
	scr_printf("Installing GS Mode Selector...\n");
	installGSModeSelector(interlace, mode, field, display, syncv, smode2);

	char CNF_filepath[] = "BOOT/GSM.CNF";
	char CNF_body[] = "invalid CNF\r\n";
	int CNF_size = strlen(CNF_body);


	if (!File_Exist(CNF_filepath)) { // Test if GSM.CNF is existing in BOOT folder
		if (!Create_File(CNF_filepath, CNF_body, CNF_size)) {
			// if GSM.CNF was NOT created in BOOT folder
			scr_printf("Error in %s creation!\n",CNF_filepath);
		} else {
			scr_printf("%s created successfully!\n",CNF_filepath);
			if (!saveConfig(CNF_filepath)) { // Test if defaults saved properly
				scr_printf("But saving default CNF values failed!\n");
			} else {
				scr_printf("And saving default CNF values worked.\n");
			}
		}
	} else
		scr_printf("%s already created!\n",CNF_filepath);

	if (!loadConfig(CNF_filepath)) {
		scr_printf("Error in %s opening!\n",CNF_filepath);
	} else
		scr_printf("%s opened successfully!\n",CNF_filepath);
	//NB: Possible failure causes include wrong CNF version ID


	scr_printf("----- Test for GSM.CNF (settings file) -----\n");
	scr_printf("--Option 1 =%s\n", GSM->option_1);	  
	scr_printf("--Option 2 =%d\n", GSM->option_2);	  


	/* Just for debbuging purposes */
	delay(5);
	scr_printf("Take a photo and press any button to continue!\n");
	while(!(waitAnyPadReady(), readpad(), new_pad)); //await a new button

outer_loop_restart:
	updateflag = 1;
//---------- Start of outer while loop ----------
	while (!(updateflag == 0)) {

		presets = make_display_magic_number(gs_dh, gs_dw, gs_magv, gs_magh, gs_dy, gs_dx);
		if(updateflag == 1){
			act_interlace = interlace;
			act_mode = mode;
			act_field = field;
			act_syncv = syncv;
			display = presets;
			ModdedSetGsCrtDisplayHandlerParams(interlace, mode, field, display, syncv, smode2);
			scr_printf("This apparently useless printf is needed before the following init_scr. Do not ask me why! ;-)\n");
			init_scr();
		}//ends if
			
		set_kmode();
		adapt1 = Adapt_Flag_1;
		adapt2 = Adapt_Flag_2;
		adapt3 = Adapt_Flag_3;
		dispreq = Safe_Orig_Display;
		dispcalc = Calc_Display;
		local_CSR_copy = GS_reg_CSR;
		set_umode();

		scr_clear();
					
		scr_printf("%s - by %s\n", VERSION, AUTHORS);
		scr_printf("------------------------------------------------------------------------------\n\n");
		scr_printf("GS(Graphics Synthesizer) Revision Number:%u - ID:%u\n", \
		(u8)( local_CSR_copy >> 16 & 0xFF ), \
		(u8)( local_CSR_copy >> 24 & 0xFF ));
		scr_printf("----- Active Settings ----- DISPLAYx 0x%08lx\n", display);
		scr_printf("Interlace %01d, Mode 0x%02x, Field %01d, SYNCV 0x%08lx\n", act_interlace, act_mode, act_field, act_syncv);
		scr_printf("(DX=%04d,DY=%04d,MAGH=%02d,MAGV=%01d,DW=%04d,DH=%04d)\n", \
		(u16)((display >> 00) & 0x0FFF), \
		(u16)((display >> 12) & 0x07FF), \
		(u16)((display >> 23) & 0x000F), \
		(u16)((display >> 27) & 0x0003), \
		(u16)((display >> 32) & 0x0FFF), \
		(u16)((display >> 44) & 0x07FF));
		scr_printf("----- Current Presets ----- DISPLAYx 0x%08lx\n", presets);
		scr_printf("Interlace %01d, Mode 0x%02x, Field %01d, SYNCV 0x%08lx\n", interlace, mode, field, syncv);
		scr_printf("(DX=%04d,DY=%04d,MAGH=%02d,MAGV=%01d,DW=%04d,DH=%04d)\n", \
		(u16)((presets >> 00) & 0x0FFF), \
		(u16)((presets >> 12) & 0x07FF), \
		(u16)((presets >> 23) & 0x000F), \
		(u16)((presets >> 27) & 0x0003), \
		(u16)((presets >> 32) & 0x0FFF), \
		(u16)((presets >> 44) & 0x07FF));
		scr_printf("----------PreDef Settings-----------------------------------------------------\n");
		scr_printf("[CIRCLE] + [LEFT]: NTSC 448i  @60Hz  | [CIRCLE] + [RIGHT]: HDTV  480p @60Hz \n");
		scr_printf("[TRIANG] + [LEFT]: PAL  512i  @50Hz  | [TRIANG] + [RIGHT]: HDTV  576p @60Hz \n");
		scr_printf("[CROSS]  + [LEFT]: HDTV 720p  @60Hz  | [CROSS]  + [RIGHT]: HDTV 1080i @60Hz \n");
		scr_printf("[SQUARE] + [LEFT]: VGA 1280x448 @60Hz| [SQUARE] + [RIGHT]: VGA 1280x448 @72Hz\n");
		scr_printf("[SELECT] + [LEFT]: VGA 1280x448 @75Hz| [SELECT] + [RIGHT]: VGA 1024x768 @75Hz\n");
		scr_printf("[SELECT] + [UP]  : VGA 1280x896 @60Hz\n");
		scr_printf("[CIRCLE] + [UP]  : NTSC(PS1 PAL game)| [TRIANG] + [UP]   : PAL(PS1 NTSC game)\n");
		scr_printf("----------Preset Commands-----------------------------------------------------\n");
		scr_printf("[Left analog stick]: DX and DY    | [Right analog stick]: DW and DH  \n");
		scr_printf("      Combos: +[L1]: Step 1 | +[none]: Step 4 | +[R1]:Step 16        \n");
		scr_printf("                     [DOWN]:Activate Presets\n");   
		scr_printf("----------Other Commands------------------------------------------------------\n");
		scr_printf("[L1] + [R1] + [CROSS] : DISPLAYx Adaption fix == %s\n", on_off_s[adapt1]);
		scr_printf("[L1] + [R1] + [CIRCLE]: Separate  SMODE2 fix  == %s\n", on_off_s[adapt2]);
		scr_printf("[L1] + [R1] + [TRIANG]: Separate  SYNCV  fix  == %s\n", on_off_s[adapt3]);
		scr_printf("[L2] + [R2]: Save CNF File\n");
		scr_printf("[R2] + [LEFT] / [RIGHT] : User Index = %d\n", user_index);
		scr_printf("[R2] + [Down] / [Up]    : Presets from / to user_mode[%d] = ", user_index);
		if(user_vmode[user_index].mode == 0)
			scr_printf("unused\n");
		else
			scr_printf("0x%02x\n", user_vmode[user_index].mode);
		scr_printf("----------Exit Commands-------------------------------------------------------\n");
		scr_printf("    [START]+[<]:PS2 BROWSER | [START]+[^]:Dev1 Boot | [START]+[>]:HDLoader\n");

		delay(5);
		
		updateflag = -1;
		
//---------- Start of inner while loop ----------
		while (updateflag == -1) {

			while(!(waitAnyPadReady(), readpad(), new_pad)); //await a new button
			retval = paddata;

			if(paddata & PAD_JOY){
				if(paddata & PAD_L1)
					joy_step = 1;
				else if(paddata & PAD_R1)
					joy_step = 16;
				else
					joy_step = 4;
			}else
				joy_step = 0;

			/* Usual pre-def settings */
			if((retval & PAD_LEFT)&&(retval & PAD_CIRCLE)) {
				interlace = 0x01;			// Interlaced
				mode = 0x02;					// NTSC 448i
				field = 0x00;					// Field mode (Read Every Other Line)
				gs_dx=700; gs_dy=46; gs_magh=3; gs_magv=0;
				gs_dw = 2560; gs_dh = 448;
				syncv = 0;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_CIRCLE)) {
				interlace = 0x00;			// Non-interlaced
				mode = 0x50;					// HDTV-480p
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=272; gs_dy=48; gs_magh=1; gs_magv=0;
				gs_dw = 1360; gs_dh = 480;
				syncv = 0x00C78C0001E00006;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_TRIANGLE)) {
				interlace = 0x01;			// Interlaced
				mode = 0x03;					// PAL 512i
				field = 0x00;					// Field mode (Read Every Other Line)
				gs_dx=720; gs_dy=68; gs_magh=3; gs_magv=0;
				gs_dw = 2560; gs_dh = 512;
				syncv = 0;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_TRIANGLE)) {
				interlace = 0x00;			// Non-interlaced
				mode = 0x53;					// HDTV-576p
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=320; gs_dy=64; gs_magh=1; gs_magv=0;
				gs_dw = 1312; gs_dh = 576;
				syncv = 0;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_CROSS))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x52;					// HDTV-720P
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=304; gs_dy=44; gs_magh=1; gs_magv=0;
				gs_dw = 1280; gs_dh = 720;
				syncv = 0x00AB400001400005;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_CROSS))	{
				interlace = 0x01;			// Interlaced
				mode = 0x51;					// HDTV-1080i
				field = 0x00;					// Field mode (Read Every Other Line)
				gs_dx=248; gs_dy=48; gs_magh=2; gs_magv=1;
				gs_dw = 1920; gs_dh = 1080;
				syncv = 0x0150E00201C00005;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_SQUARE))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x1A;					// VGA-640@60
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=272; gs_dy=34; gs_magh=1; gs_magv=0;
				gs_dw = 1280; gs_dh = 448;
//				syncv = 0x004F80000210000A; //RA NB: This does not work on my HDTV VGA
				syncv = 0x004780000210000A;	//RA NB: This does work on HDTV VGA
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_SQUARE))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x1B;					// VGA-640@72
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=312; gs_dy=18; gs_magh=1; gs_magv=0;
				gs_dw = 1280; gs_dh = 448;
//				gs_dx=360; gs_dy=30; gs_magh=1; gs_magv=0;
//				gs_dw = 1152; gs_dh = 480;
				syncv = 0;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_SELECT))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x1C;					// VGA-640@75
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=352; gs_dy=18; gs_magh=1; gs_magv=0;
				gs_dw = 1280; gs_dh = 448;
//				gs_dx=372; gs_dy=22; gs_magh=1; gs_magv=0;
//				gs_dw = 1152; gs_dh = 480;
				syncv = 0x0067800001000001;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_SELECT))	{
				interlace = 0x00;			// Non-interlaced
				mode = 0x3D;					// VGA 768p @ 75Hz
				field = 0x01;					// Frame mode (Read Every Line)
				gs_dx=256; gs_dy=30; gs_magh=0; gs_magv=0;
				gs_dw = 1024; gs_dh = 768;
				syncv = 0x006C000001C00001;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_UP)&&(retval & PAD_SELECT))	{
				interlace = 0x01;			// Interlaced
				mode = 0x1A;					// VGA 640@ 60Hz
				field = 0x00;					// Field mode (Read Every Other Line)
				gs_dx=275; gs_dy=40; gs_magh=1; gs_magv=1;
				gs_dw = 1280; gs_dh = 896;
				syncv = 0x004F80000210000A;
				smode2 = (field<<1)|interlace;
				updateflag = 1;
			}
			else if((retval & PAD_UP)&&(retval & PAD_CIRCLE))	{
				//This vmode is intended for use in enforcing NTSC for a PAL PS1 game
				interlace = 0x00;			//PS1-changed from Non-interlaced to Interlaced
				mode = 0x50;					//PS1-changed from HDTV-480p to NTSC
				field = 0x01;					//PS1-changed from Frame mode to field mode (?)
				gs_dx=736; gs_dy=12; gs_magh=1; gs_magv=0;
				gs_dw = 2560; gs_dh = 256;
				syncv = 0x00C78C0001E00006;
				smode2 = (field<<1)|interlace;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_UP)&&(retval & PAD_TRIANGLE))	{
				//This vmode is intended for use in enforcing PAL for an NTSC PS1 game
				interlace = 0x00;			//PS1-changed from Non-interlaced to Interlaced
				mode = 0x53;					//PS1-changed from HDTV-576p to PAL
				field = 0x01;					//PS1-changed from Frame mode to field mode (?)
				gs_dx=756; gs_dy=23; gs_magh=1; gs_magv=0;
				gs_dw = 2560; gs_dh = 256;
				syncv = 0;
				smode2 = (field<<1)|interlace;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if(retval == PAD_DOWN)	{
				updateflag = 1; //exit inner loop and activate presets
			}
			else if(paddata & (PAD_L3_H0 | PAD_L3_H1))	{
				gs_dx += (1-2*(!(paddata & PAD_L3_H1)))*joy_step;
				gs_dx &= (1<<12)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if(paddata & (PAD_L3_V0 | PAD_L3_V1))	{
				gs_dy += (1-2*(!(paddata & PAD_L3_V1)))*joy_step;
				gs_dy &= (1<<11)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if(paddata & (PAD_R3_H0 | PAD_R3_H1))	{
				gs_dw += (1-2*(!(paddata & PAD_R3_H1)))*joy_step;
				gs_dw &= (1<<12)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if(paddata & (PAD_R3_V0 | PAD_R3_V1))	{
				gs_dh += (1-2*(!(paddata & PAD_R3_V1)))*joy_step;
				gs_dh &= (1<<11)-1;
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_L1)&&(retval & PAD_R1)&&(retval & PAD_CROSS))	{
				set_kmode();
				Adapt_Flag_1 ^= 1;			//Toggle "DISPLAYx Adaption" flag ON/OFF
  			set_umode();
				updateflag = 1;
			}
			else if((retval & PAD_L1)&&(retval & PAD_R1)&&(retval & PAD_CIRCLE))	{
				set_kmode();
				Adapt_Flag_2 ^= 1;			//Toggle "Separate SMODE2" flag ON/OFF
  			set_umode();
				updateflag = 1;
			}
			else if((retval & PAD_L1)&&(retval & PAD_R1)&&(retval & PAD_TRIANGLE))	{
				set_kmode();
				Adapt_Flag_3 ^= 1;			//Toggle "Separate SYNCV" flag ON/OFF
  			set_umode();
				updateflag = 1;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_R2))	{
				user_index = (user_index - 1) & (user_vmode_slots-1);
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_R2))	{
				user_index = (user_index + 1) & (user_vmode_slots-1);
				updateflag = 2; //Exit inner loop but don't update GS yet
			}
			else if((retval & PAD_UP)&&(retval & PAD_R2))	{
				user_vmode[user_index].interlace = interlace;
				user_vmode[user_index].mode = mode;
				user_vmode[user_index].field = field;
				user_vmode[user_index].display = presets;
				user_vmode[user_index].syncv = syncv;
				user_vmode[user_index].smode2 = smode2;
				updateflag = 2;
			}
			else if((retval & PAD_DOWN)&&(retval & PAD_R2))	{
				if(user_vmode[user_index].mode > 0) {
					interlace = user_vmode[user_index].interlace;
					mode = user_vmode[user_index].mode;
					field = user_vmode[user_index].field;
					presets = user_vmode[user_index].display;
					syncv = user_vmode[user_index].syncv;
					smode2 = user_vmode[user_index].smode2;
					if(smode2 == 0) { //if smode2 is zero (always for CNF from v0.22)
						smode2 = (field<<1)|interlace;	//calculate smode2 form other parts
					}
					gs_dx		=	(u32)((presets >> 00) & 0x0FFF);
					gs_dy		=	(u32)((presets >> 12) & 0x07FF);
					gs_magh	=	(u32)((presets >> 23) & 0x000F);
					gs_magv	=	(u32)((presets >> 27) & 0x0003);
					gs_dw		=	(u32)((presets >> 32) & 0x0FFF)+1;
					gs_dh		=	(u32)((presets >> 44) & 0x07FF)+1;
					updateflag = 2;
				}
			}
			else if((retval & PAD_L2)&&(retval & PAD_R2))	{
				saveConfig(CNF_filepath);
				updateflag = 2;
			}
			else if((retval & PAD_LEFT)&&(retval & PAD_START))	{
				updateflag = 0;
			}
			else if((retval & PAD_UP)&&(retval & PAD_START))	{
				updateflag = 0;
			}
			else if((retval & PAD_RIGHT)&&(retval & PAD_START))	{
				updateflag = 0;
			}
		}	//---------- End of inner while loop ----------
	}	//---------- End of outer while loop ----------

	if((retval & PAD_LEFT)&&(retval & PAD_START))	{
		CleanUp();
		scr_printf("\nExiting to PS2 Browser...\n");
		delay(5);
		__asm__ __volatile__(
		".set noreorder\n"
		"li $3, 0x04\n"
		"syscall\n"
		"nop\n"
		".set reorder\n"
		);
	}
	else if((retval & PAD_UP)&&(retval & PAD_START))	{
		strcpy(elf_path, BOOT_ELF);
		goto	launch_elf;
	}
	else if((retval & PAD_RIGHT)&&(retval & PAD_START))	{
		strcpy(elf_path, HDL_ELF);
launch_elf:
		if(File_Exist(elf_path)){
			char elf_path2[0x40];

			sprintf(elf_path2, "mc0:%s", elf_path);
			elf_path2[2] += MC_index;
			CleanUp();
			scr_printf("\nLoading %s...\n", elf_path);
			delay(5);
			RunLoaderElf(elf_path2, party);	// Run ELF
		}
		scr_printf("\nCan't run %s...\n", elf_path);  // Failed to run ELF
		goto outer_loop_restart;
	}
   
   	SleepThread();					// Should never get here
   
   	return 0;
}

// End of program, leave a blank line afterwards to let 'make' line-command EE-GCC (Emotion Engine - Gnu Compiler Collection) happy ;-)

